"""ForEach iteration node."""

from collections.abc import Sequence
from functools import cached_property
from typing import Generic, Literal, Self, Type, TypeVar

from overrides import override
from pydantic import Field
from workflow_engine import (
    Context,
    Data,
    DataValue,
    Edge,
    Empty,
    InputEdge,
    IntegerValue,
    JSONValue,
    Node,
    OutputEdge,
    Params,
    SequenceValue,
    UserException,
    Value,
    ValueType,
    Workflow,
)
from workflow_engine.core import NodeTypeInfo as WENodeTypeInfo
from workflow_engine.core.values import build_data_type

from ..field import FieldInfo, FieldType
from ..node_base import AceTeamNode
from ..node_info import NodeTypeInfo
from ..workflow import AceTeamWorkflow

V = TypeVar("V", bound=Value)
D = TypeVar("D", bound=Data)


class SequenceParams(Params):
    length: IntegerValue


class SequenceData(Data, Generic[V]):
    sequence: SequenceValue[V]


class GatherSequenceNode(Node[Data, SequenceData, SequenceParams]):
    """Creates a sequence from indexed inputs."""

    TYPE_INFO = WENodeTypeInfo.from_parameter_type(
        name="GatherSequence",
        display_name="Gather Sequence",
        description="Creates a sequence from indexed inputs.",
        version="0.4.0",
        parameter_type=SequenceParams,
    )

    type: Literal["GatherSequence"] = "GatherSequence"  # pyright: ignore[reportIncompatibleVariableOverride]
    element_type: ValueType = Field(default=Value, exclude=True)

    def key(self, index: int) -> str:
        return f"element_{index}"

    @property
    def keys(self) -> list[str]:
        N = self.params.length.root
        return [self.key(i) for i in range(N)]

    @property
    def input_type(self) -> Type[Data]:
        return build_data_type(
            "GatherSequenceInput",
            {key: (self.element_type, True) for key in self.keys},
        )

    @property
    def output_type(self) -> Type[SequenceData]:
        return SequenceData[self.element_type]

    @override
    async def run(self, context: Context, input: Data) -> SequenceData:
        input_dict = input.to_dict()
        return self.output_type(
            sequence=SequenceValue[self.element_type](root=[input_dict[key] for key in self.keys])
        )

    @classmethod
    def from_length(cls, id: str, length: int, element_type: ValueType = Value) -> Self:
        return cls(
            id=id,
            params=SequenceParams(length=IntegerValue(root=length)),
            element_type=element_type,
        )


class ExpandSequenceNode(Node[SequenceData, Data, SequenceParams]):
    """Extracts sequence elements into indexed outputs."""

    TYPE_INFO = WENodeTypeInfo.from_parameter_type(
        name="ExpandSequence",
        display_name="Expand Sequence",
        description="Extracts sequence elements into indexed outputs.",
        version="0.4.0",
        parameter_type=SequenceParams,
    )

    type: Literal["ExpandSequence"] = "ExpandSequence"  # pyright: ignore[reportIncompatibleVariableOverride]
    element_type: ValueType = Value

    def key(self, index: int) -> str:
        return f"element_{index}"

    @property
    def keys(self) -> list[str]:
        N = self.params.length.root
        return [self.key(i) for i in range(N)]

    @property
    def input_type(self) -> Type[SequenceData]:
        return SequenceData[self.element_type]

    @property
    def output_type(self) -> Type[Data]:
        return build_data_type(
            "ExpandSequenceOutput",
            {key: (self.element_type, True) for key in self.keys},
        )

    @override
    async def run(self, context: Context, input: SequenceData) -> Data:
        N = self.params.length.root
        assert len(input.sequence) == N, (
            f"Expected sequence of length {N}, but got {len(input.sequence)}"
        )
        return self.output_type(**{self.key(i): input.sequence[i] for i in range(N)})

    @classmethod
    def from_length(cls, id: str, length: int, element_type: ValueType = Value) -> Self:
        return cls(
            id=id,
            params=SequenceParams(length=IntegerValue(root=length)),
            element_type=element_type,
        )


class NestedData(Data, Generic[D]):
    data: DataValue[D]


class GatherDataNode(Node[Data, NestedData, Empty]):
    """Gathers a data object into a nested `data` field."""

    TYPE_INFO = WENodeTypeInfo.from_parameter_type(
        name="GatherData",
        display_name="Gather Data",
        description="Gathers a data object into a nested data field.",
        version="0.4.0",
        parameter_type=Empty,
    )

    type: Literal["GatherData"] = "GatherData"  # pyright: ignore[reportIncompatibleVariableOverride]
    data_type: Type[Data] = Field(default=Data, exclude=True)

    @property
    def input_type(self) -> Type[Data]:
        return self.data_type

    @property
    def output_type(self) -> Type[NestedData]:
        return NestedData[self.data_type]

    @override
    async def run(self, context: Context, input: Data) -> NestedData:
        return NestedData[self.data_type](data=DataValue[self.data_type](root=input))

    @classmethod
    def from_data_type(cls, id: str, data_type: Type[Data]) -> Self:
        return cls(id=id, params=Empty(), data_type=data_type)


class ExpandDataNode(Node[NestedData, Data, Empty]):
    """Expands a nested `data` object into its fields."""

    TYPE_INFO = WENodeTypeInfo.from_parameter_type(
        name="ExpandData",
        display_name="Expand Data",
        description="Expands a nested data object into its fields.",
        version="0.4.0",
        parameter_type=Empty,
    )

    type: Literal["ExpandData"] = "ExpandData"  # pyright: ignore[reportIncompatibleVariableOverride]
    data_type: Type[Data] = Field(default=Data, exclude=True)

    @property
    def input_type(self) -> Type[NestedData]:
        return NestedData[self.data_type]

    @property
    def output_type(self) -> Type[Data]:
        return self.data_type

    @override
    async def run(self, context: Context, input: NestedData) -> Data:
        return input.data.root

    @classmethod
    def from_data_type(cls, id: str, data_type: Type[Data]) -> Self:
        return cls(id=id, params=Empty(), data_type=data_type)


class JsonToSequenceInput(Data):
    sequence: JSONValue


class JsonToSequenceOutput(Data, Generic[V]):
    sequence: SequenceValue[V]


class JsonToSequenceNode(Node[JsonToSequenceInput, JsonToSequenceOutput, Empty]):
    """Converts a JSON array into a typed sequence."""

    TYPE_INFO = WENodeTypeInfo.from_parameter_type(
        name="JsonToSequence",
        display_name="Json to Sequence",
        description="Converts a JSON array into a typed sequence.",
        version="0.4.0",
        parameter_type=Empty,
    )

    type: Literal["JsonToSequence"] = "JsonToSequence"
    element_type: ValueType = Value
    data_type: Type[Data] = Field(default=Data, exclude=True)

    @property
    def input_type(self) -> Type[JsonToSequenceInput]:
        return JsonToSequenceInput

    @property
    def output_type(self) -> Type[JsonToSequenceOutput]:
        return JsonToSequenceOutput[DataValue[self.data_type]]

    @override
    async def run(self, context: Context, input: JsonToSequenceInput) -> JsonToSequenceOutput:
        items = input.sequence.root
        if not isinstance(items, Sequence):
            raise UserException(f"Expected a JSON sequence, but got {type(items)}: {items}")
        typed_items: list[DataValue] = []
        for item in items:
            typed_items.append(DataValue[self.data_type](root=self.data_type.model_validate(item)))
        return self.output_type(
            sequence=SequenceValue[DataValue[self.data_type]](root=typed_items)
        )

    @classmethod
    def from_data_type(cls, id: str, data_type: Type[Data]) -> Self:
        return cls(id=id, params=Empty(), data_type=data_type)


class ForEachParams(Params):
    workflow: JSONValue = Field(default_factory=lambda: JSONValue(None))

    @property
    def ace_workflow(self) -> AceTeamWorkflow:
        return AceTeamWorkflow.model_validate(self.workflow.root)


class ForEachNode(AceTeamNode[SequenceData, SequenceData, ForEachParams]):
    """Executes a nested workflow for each element in the input sequence."""

    TYPE_INFO = WENodeTypeInfo.from_parameter_type(
        name="ForEach",
        display_name="For Each",
        description="Executes a nested workflow for each element in the input sequence.",
        version="0.4.0",
        parameter_type=ForEachParams,
    )

    type: Literal["ForEach"] = "ForEach"  # pyright: ignore[reportIncompatibleVariableOverride]

    @classmethod
    @override
    def type_info(cls) -> NodeTypeInfo:
        return NodeTypeInfo(
            type="ForEach",
            display_name="For Each",
            description="Executes a nested workflow for each element in the input sequence.",
            params=(
                FieldInfo(
                    name="workflow",
                    display_name="Workflow",
                    description="Workflow to run for each element.",
                    type=FieldType.WORKFLOW,
                ),
            ),
        )

    @cached_property
    def input_fields_info(self) -> Sequence[FieldInfo]:
        return (
            FieldInfo(
                name="sequence",
                display_name="Sequence",
                description="Sequence of items to process.",
                type=FieldType.JSON,
            ),
        )

    @cached_property
    def output_fields_info(self) -> Sequence[FieldInfo]:
        return (
            FieldInfo(
                name="sequence",
                display_name="Sequence",
                description="Processed sequence.",
                type=FieldType.JSON,
            ),
        )

    @override
    async def run(self, context: Context, input: Data) -> Workflow:
        seq = getattr(input, "sequence")
        N = len(seq.root)
        workflow = self.params.ace_workflow

        nodes: list[Node] = []
        edges: list[Edge] = []

        json_to_sequence = JsonToSequenceNode.from_data_type(
            id="json_to_sequence", data_type=workflow.input_type
        )
        expand = ExpandSequenceNode.from_length(
            id="expand", length=N, element_type=DataValue[workflow.input_type]
        )
        gather = GatherSequenceNode.from_length(
            id="gather", length=N, element_type=DataValue[workflow.output_type]
        )
        nodes.append(json_to_sequence)
        nodes.append(expand)
        nodes.append(gather)

        for i in range(N):
            namespace = f"element_{i}"
            input_adapter = ExpandDataNode.from_data_type(
                id="input_adapter", data_type=workflow.input_type
            ).with_namespace(namespace)
            item_workflow = workflow.with_namespace(namespace)
            output_adapter = GatherDataNode.from_data_type(
                id="output_adapter", data_type=workflow.output_type
            ).with_namespace(namespace)

            nodes.append(input_adapter)
            nodes.extend(item_workflow.nodes)
            nodes.append(output_adapter)

            edges.append(
                Edge.from_nodes(
                    source=expand, source_key=expand.key(i),
                    target=input_adapter, target_key="data",
                )
            )
            for input_edge in item_workflow.input_edges:
                edges.append(
                    Edge(
                        source_id=input_adapter.id, source_key=input_edge.input_key,
                        target_id=input_edge.target_id, target_key=input_edge.target_key,
                    )
                )
            edges.extend(item_workflow.edges)
            for output_edge in item_workflow.output_edges:
                edges.append(
                    Edge(
                        source_id=output_edge.source_id, source_key=output_edge.source_key,
                        target_id=output_adapter.id, target_key=output_edge.output_key,
                    )
                )
            edges.append(
                Edge.from_nodes(
                    source=output_adapter, source_key="data",
                    target=gather, target_key=gather.key(i),
                )
            )

        edges.append(
            Edge.from_nodes(
                source=json_to_sequence, source_key="sequence",
                target=expand, target_key="sequence",
            )
        )

        return Workflow(
            nodes=nodes,
            edges=edges,
            input_edges=[
                InputEdge.from_node(
                    input_key="sequence", target=json_to_sequence, target_key="sequence"
                )
            ],
            output_edges=[
                OutputEdge.from_node(
                    source=gather, source_key="sequence", output_key="sequence"
                )
            ],
        )


__all__ = [
    "ForEachNode",
    "ForEachParams",
]



"""AceTeam node base class with auto-registration."""

from collections.abc import Mapping, Sequence
from functools import cached_property
from typing import Any, Generic, Type, TypeVar

from workflow_engine import Data, DataMapping, Node, Params
from workflow_engine.core import NodeTypeInfo as WENodeTypeInfo

from .field import FieldInfo, build_pydantic_model
from .node_info import NodeTypeInfo

Input_contra = TypeVar("Input_contra", bound=Data, contravariant=True)
Output_co = TypeVar("Output_co", bound=Data, covariant=True)
Params_co = TypeVar("Params_co", bound=Params, covariant=True)


class AceTeamNode(
    Node[Input_contra, Output_co, Params_co],
    Generic[Input_contra, Output_co, Params_co],
):
    """
    An AceTeam-specific node with field information for dynamic input/output types.
    """

    TYPE_INFO = WENodeTypeInfo.from_parameter_type(
        name="AceTeamNode",
        display_name="AceTeam Node",
        version="0.4.0",
        parameter_type=Params,
    )

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if cls.__name__.split("[")[0] != "AceTeamNode":
            assert cls not in aceteam_nodes
            aceteam_nodes.append(cls)

    @classmethod
    def type_info(cls) -> NodeTypeInfo:
        """Fixed information about the node and its properties."""
        raise NotImplementedError("Subclasses must implement this method")

    @cached_property
    def input_fields_info(self) -> Sequence[FieldInfo]:
        return ()

    @cached_property
    def output_fields_info(self) -> Sequence[FieldInfo]:
        return ()

    @cached_property
    def input_type(self) -> Type[Input_contra]:  # type: ignore
        return build_pydantic_model(self.input_fields_info)  # type: ignore

    @cached_property
    def output_type(self) -> Type[Output_co]:
        return build_pydantic_model(self.output_fields_info)  # type: ignore

    def parse_input(self, input: Mapping[str, Any]) -> DataMapping:
        return {
            field.name: field.python_type.model_validate(input[field.name])
            for field in self.input_fields_info
            if field.name in input
        }

    def parse_output(self, output: Mapping[str, Any]) -> DataMapping:
        return {
            field.name: field.python_type.model_validate(output[field.name])
            for field in self.output_fields_info
            if field.name in output
        }


aceteam_nodes: list[type[AceTeamNode]] = []


__all__ = [
    "AceTeamNode",
    "aceteam_nodes",
]



"""API Call node - HTTP requests with Jinja templating."""

import json
import logging
from collections.abc import Sequence
from functools import cached_property
from typing import Any, Literal

import httpx
from overrides import override
from pydantic import Field
from workflow_engine import (
    BooleanValue,
    Context,
    Data,
    File,
    FloatValue,
    IntegerValue,
    JSONValue,
    Params,
    SequenceValue,
    StringMapValue,
    StringValue,
    UserException,
    Value,
)
from workflow_engine.core import NodeTypeInfo as WENodeTypeInfo
from workflow_engine.files import JSONFileValue, JSONLinesFileValue, TextFileValue

from ..field import FieldInfo, FieldInfoValue, FieldType
from ..node_base import AceTeamNode
from ..node_info import NodeTypeInfo
from ..utils import format_jinja

logger = logging.getLogger(__name__)


class APICallParams(Params):
    url: StringValue
    method: StringValue = Field(default_factory=lambda: StringValue("GET"))
    headers: StringMapValue[StringValue] = Field(
        default_factory=lambda: StringMapValue[StringValue]({})
    )
    body_template: StringValue = Field(default_factory=lambda: StringValue(""))
    parameters: SequenceValue[FieldInfoValue] = Field(
        default_factory=lambda: SequenceValue[FieldInfoValue]([])
    )
    timeout: FloatValue = Field(default_factory=lambda: FloatValue(30.0))
    output_file: BooleanValue = Field(default_factory=lambda: BooleanValue(False))


class APICallOutput(Data):
    status_code: IntegerValue
    response: JSONValue
    headers: StringMapValue[StringValue]


class APICallTextFileOutput(Data):
    status_code: IntegerValue
    response: TextFileValue
    headers: StringMapValue[StringValue]


class APICallJSONFileOutput(Data):
    status_code: IntegerValue
    response: JSONFileValue
    headers: StringMapValue[StringValue]


class APICallNode(
    AceTeamNode[
        Data,
        APICallOutput | APICallTextFileOutput | APICallJSONFileOutput,
        APICallParams,
    ]
):
    """Node that makes HTTP API calls with Jinja templating for URL and request body."""

    TYPE_INFO = WENodeTypeInfo.from_parameter_type(
        name="APICall",
        display_name="API Call",
        description="Makes HTTP API calls with Jinja templating for URL and request body.",
        version="0.4.0",
        parameter_type=APICallParams,
    )

    type: Literal["APICall"] = "APICall"

    @classmethod
    @override
    def type_info(cls) -> NodeTypeInfo:
        return NodeTypeInfo(
            type="APICall",
            display_name="API Call",
            description="Makes HTTP API calls with Jinja templating for URL and request body.",
            params=(
                FieldInfo(
                    name="url",
                    display_name="URL Template",
                    description="The API endpoint URL.",
                    type=FieldType.SHORT_TEXT,
                ),
                FieldInfo(
                    name="method",
                    display_name="HTTP Method",
                    description="The HTTP method to use.",
                    type=FieldType.SHORT_TEXT,
                    default="GET",
                ),
                FieldInfo(
                    name="headers",
                    display_name="Headers",
                    description="HTTP headers to include in the request.",
                    type=FieldType.JSON,
                    default={},
                ),
                FieldInfo(
                    name="body_template",
                    display_name="Request Body Template",
                    description="The request body template.",
                    type=FieldType.LONG_TEXT,
                    default="",
                ),
                FieldInfo(
                    name="timeout",
                    display_name="Timeout (seconds)",
                    description="Request timeout in seconds.",
                    type=FieldType.NUMBER,
                    default=30.0,
                    min=1.0,
                    max=300.0,
                    step=1.0,
                ),
                FieldInfo(
                    name="output_file",
                    display_name="Output to File",
                    description="Whether to save the response to a file.",
                    type=FieldType.BOOLEAN,
                    default=False,
                ),
                FieldInfo(
                    name="parameters",
                    display_name="Template Parameters",
                    description="Parameters used in URL and body templates.",
                    type=FieldType.FIELD_LIST,
                    default=(),
                ),
            ),
        )

    @cached_property
    def supports_json_output(self) -> bool:
        content_type = self.params.headers.get("Accept")
        return content_type is not None and "application/json" in content_type

    @cached_property
    def input_fields_info(self) -> Sequence[FieldInfo]:
        return tuple(field.root for field in self.params.parameters)

    @cached_property
    def output_fields_info(self) -> Sequence[FieldInfo]:
        if self.params.output_file:
            return (
                FieldInfo(
                    name="status_code",
                    display_name="Status Code",
                    description="HTTP response status code.",
                    type=FieldType.NUMBER,
                ),
                FieldInfo(
                    name="response",
                    display_name="Response File",
                    description="File containing the API response.",
                    type=FieldType.FILE_JSON
                    if self.supports_json_output
                    else FieldType.FILE_TEXT,
                ),
                FieldInfo(
                    name="headers",
                    display_name="Response Headers",
                    description="HTTP response headers.",
                    type=FieldType.JSON,
                ),
            )
        else:
            return (
                FieldInfo(
                    name="status_code",
                    display_name="Status Code",
                    description="HTTP response status code.",
                    type=FieldType.NUMBER,
                ),
                FieldInfo(
                    name="response",
                    display_name="Response",
                    description="The API response data.",
                    type=FieldType.JSON,
                ),
                FieldInfo(
                    name="headers",
                    display_name="Response Headers",
                    description="HTTP response headers.",
                    type=FieldType.JSON,
                ),
            )

    @property
    def output_type(self):
        if self.params.output_file:
            if self.supports_json_output:
                return APICallJSONFileOutput
            else:
                return APICallTextFileOutput
        else:
            return APICallOutput

    async def _expand_parameter_value(self, context: Context, value: Value) -> Any:
        """Expand data files into their content for templating."""
        assert isinstance(value, Value)
        if isinstance(value, JSONFileValue):
            try:
                return await value.read_data(context)
            except Exception as e:
                raise UserException(f"Failed to read JSON file '{value.path}'.") from e
        elif isinstance(value, JSONLinesFileValue):
            try:
                return await value.read_data(context)
            except Exception as e:
                raise UserException(f"Failed to read JSON lines file '{value.path}'.") from e
        elif isinstance(value, TextFileValue):
            try:
                return await value.read_text(context)
            except Exception as e:
                raise UserException(f"Failed to read text file '{value.path}'.") from e
        return value.root

    def _is_json_response(self, response: httpx.Response) -> bool:
        content_type = response.headers.get("content-type", "")
        assert isinstance(content_type, str)
        content_type = content_type.lower()
        return "application/json" in content_type or "text/json" in content_type

    def _parse_response_data(self, response: httpx.Response) -> Any:
        try:
            if self._is_json_response(response):
                return response.json()
            else:
                return response.text
        except json.JSONDecodeError:
            try:
                return response.text
            except Exception:
                return str(response.content)
        except Exception:
            return str(response.content)

    @override
    async def run(
        self, context: Context, input: Data
    ) -> APICallOutput | APICallTextFileOutput | APICallJSONFileOutput:
        # Prepare template parameters
        parameters: dict[str, Any] = {}
        if len(self.params.parameters) > 0:
            for field in self.params.parameters:
                if hasattr(input, field.name):
                    parameters[field.name] = await self._expand_parameter_value(
                        context, getattr(input, field.name)
                    )
                elif field.default is not None:
                    parameters[field.name] = field.default
                else:
                    raise UserException(
                        f"Parameter {field.name} is not set and has no default value."
                    )

        # Format URL with Jinja templating
        try:
            url_template = self.params.url.root
            url = format_jinja(url_template, parameters)
        except Exception as e:
            raise UserException(f"Failed to format URL template: {e}") from e

        # Format request body with Jinja templating
        request_body = None
        body_template = self.params.body_template.root
        if body_template.strip():
            try:
                body_text = format_jinja(body_template, parameters)
                if body_text.strip().startswith(("{", "[")):
                    try:
                        request_body = json.loads(body_text)
                    except json.JSONDecodeError:
                        request_body = body_text
                else:
                    request_body = body_text
            except Exception as e:
                raise UserException(f"Failed to format request body template: {e}") from e

        # Prepare headers
        headers = {k: v.root for k, v in self.params.headers.items()}

        if request_body is not None and isinstance(request_body, (dict, list)):
            if "content-type" not in {k.lower() for k in headers.keys()}:
                headers["Content-Type"] = "application/json"

        logger.info(f"Making {self.params.method} request to: {url}")

        # Make the HTTP request
        try:
            timeout = self.params.timeout.root
            method = self.params.method.root
            async with httpx.AsyncClient(timeout=timeout) as client:
                if isinstance(request_body, (dict, list)):
                    response = await client.request(
                        method=method, url=url, headers=headers, json=request_body
                    )
                elif request_body is not None:
                    response = await client.request(
                        method=method, url=url, headers=headers, content=request_body
                    )
                else:
                    response = await client.request(method=method, url=url, headers=headers)
        except httpx.TimeoutException as e:
            raise UserException(
                f"Request timed out after {self.params.timeout} seconds."
            ) from e
        except httpx.RequestError as e:
            raise UserException(f"Request failed: {e}") from e
        except Exception as e:
            raise UserException(f"Unexpected error during API call: {e}") from e

        # Parse response
        status_code = IntegerValue(response.status_code)
        response_data = self._parse_response_data(response)
        response_headers = StringMapValue[StringValue](
            {k: StringValue(v) for k, v in response.headers.items()}
        )

        # Handle file output
        if self.params.output_file:
            if self._is_json_response(response) and isinstance(response_data, (dict, list)):
                response_file = JSONFileValue(File(path="api_response.json"))
                response_file = await response_file.write_data(context, response_data)
                return APICallJSONFileOutput(
                    status_code=status_code, response=response_file, headers=response_headers
                )
            else:
                response_file = TextFileValue(File(path="api_response.txt"))
                response_text = (
                    response_data if isinstance(response_data, str) else str(response_data)
                )
                response_file = await response_file.write_text(context, response_text)
                return APICallTextFileOutput(
                    status_code=status_code, response=response_file, headers=response_headers
                )
        else:
            return APICallOutput(
                status_code=status_code, response=response_data, headers=response_headers
            )


__all__ = [
    "APICallNode",
    "APICallParams",
]



"""AceTeam workflow model."""

from collections.abc import Mapping, Sequence
from functools import cached_property
from typing import Any

from pydantic import model_validator
from workflow_engine import DataMapping, Workflow

from .field import FieldInfo
from .node_base import AceTeamNode


class AceTeamWorkflow(Workflow):
    nodes: Sequence[AceTeamNode]
    inputs: Sequence[FieldInfo]
    outputs: Sequence[FieldInfo]

    @cached_property
    def inputs_by_key(self) -> Mapping[str, FieldInfo]:
        inputs_by_key: dict[str, FieldInfo] = {}
        for field in self.inputs:
            name = field.name
            assert name not in inputs_by_key, f"Duplicate input field: {name}"
            inputs_by_key[name] = field
        return inputs_by_key

    @cached_property
    def outputs_by_key(self) -> Mapping[str, FieldInfo]:
        outputs_by_key: dict[str, FieldInfo] = {}
        for field in self.outputs:
            name = field.name
            assert name not in outputs_by_key, f"Duplicate output field: {name}"
            outputs_by_key[name] = field
        return outputs_by_key

    @model_validator(mode="after")
    def _validate_input_fields(self):
        for input_edge in self.input_edges:
            input_field = self.inputs_by_key[input_edge.input_key]
            input_type = input_field.python_type
            target = self.nodes_by_id[input_edge.target_id]
            input_edge.validate_types(input_type=input_type, target=target)
        return self

    @model_validator(mode="after")
    def _validate_output_fields(self):
        for output_edge in self.output_edges:
            output_field = self.outputs_by_key[output_edge.output_key]
            output_type = output_field.python_type
            source = self.nodes_by_id[output_edge.source_id]
            output_edge.validate_types(source=source, output_type=output_type)
        return self

    def parse_input(self, input: Mapping[str, Any]) -> DataMapping:
        return {
            field.name: field.python_type.model_validate(input[field.name])
            for field in self.inputs
            if field.name in input
        }

    def parse_output(self, output: Mapping[str, Any]) -> DataMapping:
        return {
            field.name: field.python_type.model_validate(output[field.name])
            for field in self.outputs
            if field.name in output
        }


__all__ = [
    "AceTeamWorkflow",
]



"""Comparison and logical operator nodes."""

from typing import Literal

from overrides import override
from workflow_engine import BooleanValue, Context, Data, FloatValue, Params
from workflow_engine.core import NodeTypeInfo as WENodeTypeInfo

from ..field import FieldInfo, FieldType
from ..node_base import AceTeamNode
from ..node_info import NodeTypeInfo


class ComparisonParams(Params):
    pass


class ComparisonInput(Data):
    a: FloatValue
    b: FloatValue


class ComparisonOutput(Data):
    result: BooleanValue


class LogicalParams(Params):
    pass


class LogicalInput(Data):
    a: BooleanValue
    b: BooleanValue


class NotInput(Data):
    a: BooleanValue


class LogicalOutput(Data):
    result: BooleanValue


# --- Comparison Nodes ---


class EqualNode(AceTeamNode[ComparisonInput, ComparisonOutput, ComparisonParams]):
    """Checks if a == b."""

    TYPE_INFO = WENodeTypeInfo.from_parameter_type(
        name="Equal",
        display_name="Equal",
        description="Checks if two numbers are equal (a == b).",
        version="0.4.0",
        parameter_type=ComparisonParams,
    )
    type: Literal["Equal"] = "Equal"

    @classmethod
    @override
    def type_info(cls) -> NodeTypeInfo:
        return NodeTypeInfo(
            type="Equal",
            display_name="Equal",
            description="Checks if two numbers are equal (a == b).",
            params=(),
        )

    @property
    def input_fields_info(self):
        return (
            FieldInfo(
                name="a",
                display_name="First Number",
                type=FieldType.NUMBER,
                description="First number.",
            ),
            FieldInfo(
                name="b",
                display_name="Second Number",
                type=FieldType.NUMBER,
                description="Second number.",
            ),
        )

    @property
    def output_fields_info(self):
        return (
            FieldInfo(
                name="result",
                display_name="Result",
                type=FieldType.BOOLEAN,
                description="True if a == b.",
            ),
        )

    @override
    async def run(self, context: Context, input: ComparisonInput) -> ComparisonOutput:
        return ComparisonOutput(result=BooleanValue(input.a.root == input.b.root))


class NotEqualNode(AceTeamNode[ComparisonInput, ComparisonOutput, ComparisonParams]):
    """Checks if a != b."""

    type: Literal["NotEqual"] = "NotEqual"

    @classmethod
    @override
    def type_info(cls) -> NodeTypeInfo:
        return NodeTypeInfo(
            type="NotEqual",
            display_name="Not Equal",
            description="Checks if two numbers are not equal (a != b).",
            params=(),
        )

    @property
    def input_fields_info(self):
        return (
            FieldInfo(
                name="a",
                display_name="First Number",
                type=FieldType.NUMBER,
                description="First number.",
            ),
            FieldInfo(
                name="b",
                display_name="Second Number",
                type=FieldType.NUMBER,
                description="Second number.",
            ),
        )

    @property
    def output_fields_info(self):
        return (
            FieldInfo(
                name="result",
                display_name="Result",
                type=FieldType.BOOLEAN,
                description="True if a != b.",
            ),
        )

    @override
    async def run(self, context: Context, input: ComparisonInput) -> ComparisonOutput:
        return ComparisonOutput(result=BooleanValue(input.a.root != input.b.root))


class GreaterThanNode(AceTeamNode[ComparisonInput, ComparisonOutput, ComparisonParams]):
    """Checks if a > b."""

    TYPE_INFO = WENodeTypeInfo.from_parameter_type(
        name="GreaterThan",
        display_name="Greater Than",
        description="Checks if a > b.",
        version="0.4.0",
        parameter_type=ComparisonParams,
    )
    type: Literal["GreaterThan"] = "GreaterThan"

    @classmethod
    @override
    def type_info(cls) -> NodeTypeInfo:
        return NodeTypeInfo(
            type="GreaterThan",
            display_name="Greater Than",
            description="Checks if a > b.",
            params=(),
        )

    @property
    def input_fields_info(self):
        return (
            FieldInfo(
                name="a",
                display_name="First Number",
                type=FieldType.NUMBER,
                description="First number.",
            ),
            FieldInfo(
                name="b",
                display_name="Second Number",
                type=FieldType.NUMBER,
                description="Second number.",
            ),
        )

    @property
    def output_fields_info(self):
        return (
            FieldInfo(
                name="result",
                display_name="Result",
                type=FieldType.BOOLEAN,
                description="True if a > b.",
            ),
        )

    @override
    async def run(self, context: Context, input: ComparisonInput) -> ComparisonOutput:
        return ComparisonOutput(result=BooleanValue(input.a.root > input.b.root))


class GreaterThanEqualNode(AceTeamNode[ComparisonInput, ComparisonOutput, ComparisonParams]):
    """Checks if a >= b."""

    type: Literal["GreaterThanEqual"] = "GreaterThanEqual"

    @classmethod
    @override
    def type_info(cls) -> NodeTypeInfo:
        return NodeTypeInfo(
            type="GreaterThanEqual",
            display_name="Greater Than or Equal",
            description="Checks if a >= b.",
            params=(),
        )

    @property
    def input_fields_info(self):
        return (
            FieldInfo(
                name="a",
                display_name="First Number",
                type=FieldType.NUMBER,
                description="First number.",
            ),
            FieldInfo(
                name="b",
                display_name="Second Number",
                type=FieldType.NUMBER,
                description="Second number.",
            ),
        )

    @property
    def output_fields_info(self):
        return (
            FieldInfo(
                name="result",
                display_name="Result",
                type=FieldType.BOOLEAN,
                description="True if a >= b.",
            ),
        )

    @override
    async def run(self, context: Context, input: ComparisonInput) -> ComparisonOutput:
        return ComparisonOutput(result=BooleanValue(input.a.root >= input.b.root))


class LessThanNode(AceTeamNode[ComparisonInput, ComparisonOutput, ComparisonParams]):
    """Checks if a < b."""

    TYPE_INFO = WENodeTypeInfo.from_parameter_type(
        name="LessThan",
        display_name="Less Than",
        description="Checks if a < b.",
        version="0.4.0",
        parameter_type=ComparisonParams,
    )
    type: Literal["LessThan"] = "LessThan"

    @classmethod
    @override
    def type_info(cls) -> NodeTypeInfo:
        return NodeTypeInfo(
            type="LessThan",
            display_name="Less Than",
            description="Checks if a < b.",
            params=(),
        )

    @property
    def input_fields_info(self):
        return (
            FieldInfo(
                name="a",
                display_name="First Number",
                type=FieldType.NUMBER,
                description="First number.",
            ),
            FieldInfo(
                name="b",
                display_name="Second Number",
                type=FieldType.NUMBER,
                description="Second number.",
            ),
        )

    @property
    def output_fields_info(self):
        return (
            FieldInfo(
                name="result",
                display_name="Result",
                type=FieldType.BOOLEAN,
                description="True if a < b.",
            ),
        )

    @override
    async def run(self, context: Context, input: ComparisonInput) -> ComparisonOutput:
        return ComparisonOutput(result=BooleanValue(input.a.root < input.b.root))


class LessThanEqualNode(AceTeamNode[ComparisonInput, ComparisonOutput, ComparisonParams]):
    """Checks if a <= b."""

    type: Literal["LessThanEqual"] = "LessThanEqual"

    @classmethod
    @override
    def type_info(cls) -> NodeTypeInfo:
        return NodeTypeInfo(
            type="LessThanEqual",
            display_name="Less Than or Equal",
            description="Checks if a <= b.",
            params=(),
        )

    @property
    def input_fields_info(self):
        return (
            FieldInfo(
                name="a",
                display_name="First Number",
                type=FieldType.NUMBER,
                description="First number.",
            ),
            FieldInfo(
                name="b",
                display_name="Second Number",
                type=FieldType.NUMBER,
                description="Second number.",
            ),
        )

    @property
    def output_fields_info(self):
        return (
            FieldInfo(
                name="result",
                display_name="Result",
                type=FieldType.BOOLEAN,
                description="True if a <= b.",
            ),
        )

    @override
    async def run(self, context: Context, input: ComparisonInput) -> ComparisonOutput:
        return ComparisonOutput(result=BooleanValue(input.a.root <= input.b.root))


# --- Logical Nodes ---


class AndNode(AceTeamNode[LogicalInput, LogicalOutput, LogicalParams]):
    """Logical AND (a && b)."""

    TYPE_INFO = WENodeTypeInfo.from_parameter_type(
        name="And",
        display_name="Logical AND",
        description="Logical AND (a && b).",
        version="0.4.0",
        parameter_type=LogicalParams,
    )
    type: Literal["And"] = "And"

    @classmethod
    @override
    def type_info(cls) -> NodeTypeInfo:
        return NodeTypeInfo(
            type="And",
            display_name="Logical AND",
            description="Logical AND (a && b).",
            params=(),
        )

    @property
    def input_fields_info(self):
        return (
            FieldInfo(
                name="a",
                display_name="First Boolean",
                type=FieldType.BOOLEAN,
                description="First boolean.",
            ),
            FieldInfo(
                name="b",
                display_name="Second Boolean",
                type=FieldType.BOOLEAN,
                description="Second boolean.",
            ),
        )

    @property
    def output_fields_info(self):
        return (
            FieldInfo(
                name="result", display_name="Result", type=FieldType.BOOLEAN, description="a AND b."
            ),
        )

    @override
    async def run(self, context: Context, input: LogicalInput) -> LogicalOutput:
        return LogicalOutput(result=BooleanValue(input.a.root and input.b.root))


class OrNode(AceTeamNode[LogicalInput, LogicalOutput, LogicalParams]):
    """Logical OR (a || b)."""

    type: Literal["Or"] = "Or"

    @classmethod
    @override
    def type_info(cls) -> NodeTypeInfo:
        return NodeTypeInfo(
            type="Or",
            display_name="Logical OR",
            description="Logical OR (a || b).",
            params=(),
        )

    @property
    def input_fields_info(self):
        return (
            FieldInfo(
                name="a",
                display_name="First Boolean",
                type=FieldType.BOOLEAN,
                description="First boolean.",
            ),
            FieldInfo(
                name="b",
                display_name="Second Boolean",
                type=FieldType.BOOLEAN,
                description="Second boolean.",
            ),
        )

    @property
    def output_fields_info(self):
        return (
            FieldInfo(
                name="result", display_name="Result", type=FieldType.BOOLEAN, description="a OR b."
            ),
        )

    @override
    async def run(self, context: Context, input: LogicalInput) -> LogicalOutput:
        return LogicalOutput(result=BooleanValue(input.a.root or input.b.root))


class NotNode(AceTeamNode[NotInput, LogicalOutput, LogicalParams]):
    """Logical NOT (!a)."""

    type: Literal["Not"] = "Not"

    @classmethod
    @override
    def type_info(cls) -> NodeTypeInfo:
        return NodeTypeInfo(
            type="Not",
            display_name="Logical NOT",
            description="Logical NOT (!a).",
            params=(),
        )

    @property
    def input_fields_info(self):
        return (
            FieldInfo(
                name="a",
                display_name="Boolean Value",
                type=FieldType.BOOLEAN,
                description="Value to negate.",
            ),
        )

    @property
    def output_fields_info(self):
        return (
            FieldInfo(
                name="result", display_name="Result", type=FieldType.BOOLEAN, description="NOT a."
            ),
        )

    @override
    async def run(self, context: Context, input: NotInput) -> LogicalOutput:
        return LogicalOutput(result=BooleanValue(not input.a.root))


__all__ = [
    "AndNode",
    "EqualNode",
    "GreaterThanEqualNode",
    "GreaterThanNode",
    "LessThanEqualNode",
    "LessThanNode",
    "NotEqualNode",
    "NotNode",
    "OrNode",
]



"""Text Input node - simple text source for workflows."""

from collections.abc import Sequence
from typing import Literal

from overrides import override
from pydantic import Field
from workflow_engine import Context, Data, Params, StringValue
from workflow_engine.core import NodeTypeInfo as WENodeTypeInfo

from ..field import FieldInfo, FieldType
from ..node_base import AceTeamNode
from ..node_info import NodeTypeInfo


class TextInputNodeParams(Params):
    text: StringValue = Field(
        default=StringValue(""),
        description="The text content to output",
    )


class TextInputNodeOutput(Data):
    output: StringValue


class TextInputNode(
    AceTeamNode[None, TextInputNodeOutput, TextInputNodeParams],
):
    """A simple text source node that outputs a configurable text string."""

    TYPE_INFO = WENodeTypeInfo.from_parameter_type(
        name="TextInput",
        display_name="Text Input",
        description="A simple text source.",
        version="1.0.0",
        parameter_type=TextInputNodeParams,
    )

    type: Literal["TextInput"] = "TextInput"

    @classmethod
    @override
    def type_info(cls) -> NodeTypeInfo:
        return NodeTypeInfo(
            type="TextInput",
            display_name="Text Input",
            description="A simple text source.",
            params=(
                FieldInfo(
                    name="text",
                    display_name="Text",
                    description="The text content to output",
                    type=FieldType.LONG_TEXT,
                    default="",
                ),
            ),
        )

    @property
    def input_fields_info(self) -> Sequence[FieldInfo]:
        return ()

    @property
    def output_fields_info(self) -> Sequence[FieldInfo]:
        return (
            FieldInfo(
                name="output",
                display_name="Output",
                description="The text content",
                type=FieldType.SHORT_TEXT,
            ),
        )

    @override
    async def run(self, context: Context, input: None) -> TextInputNodeOutput:
        return TextInputNodeOutput(output=self.params.text)


__all__ = [
    "TextInputNode",
    "TextInputNodeOutput",
    "TextInputNodeParams",
]



"""Data Transform node - simple data transformation."""

from collections.abc import Sequence
from typing import Literal

from overrides import override
from pydantic import Field
from workflow_engine import Context, Data, Params, StringValue
from workflow_engine.core import NodeTypeInfo as WENodeTypeInfo

from ..field import FieldInfo, FieldType
from ..node_base import AceTeamNode
from ..node_info import NodeTypeInfo


class DataTransformNodeParams(Params):
    instruction: StringValue = Field(
        default=StringValue("Pass through unchanged"),
        description="Description of the transformation to apply",
    )


class DataTransformNodeInput(Data):
    input: StringValue


class DataTransformNodeOutput(Data):
    output: StringValue


class DataTransformNode(
    AceTeamNode[DataTransformNodeInput, DataTransformNodeOutput, DataTransformNodeParams],
):
    """Processes and transforms input data based on configured instructions."""

    TYPE_INFO = WENodeTypeInfo.from_parameter_type(
        name="DataTransform",
        display_name="Transform",
        description="Processes and transforms data.",
        version="1.0.0",
        parameter_type=DataTransformNodeParams,
    )

    type: Literal["DataTransform"] = "DataTransform"

    @classmethod
    @override
    def type_info(cls) -> NodeTypeInfo:
        return NodeTypeInfo(
            type="DataTransform",
            display_name="Transform",
            description="Processes and transforms data.",
            params=(
                FieldInfo(
                    name="instruction",
                    display_name="Instruction",
                    description="Description of the transformation to apply",
                    type=FieldType.LONG_TEXT,
                    default="Pass through unchanged",
                ),
            ),
        )

    @property
    def input_fields_info(self) -> Sequence[FieldInfo]:
        return (
            FieldInfo(
                name="input",
                display_name="Input",
                description="The data to transform",
                type=FieldType.LONG_TEXT,
            ),
        )

    @property
    def output_fields_info(self) -> Sequence[FieldInfo]:
        return (
            FieldInfo(
                name="output",
                display_name="Output",
                description="The transformed data",
                type=FieldType.LONG_TEXT,
            ),
        )

    @override
    async def run(
        self, context: Context, input: DataTransformNodeInput
    ) -> DataTransformNodeOutput:
        # Pass through - transformation logic can be extended
        return DataTransformNodeOutput(output=input.input)


__all__ = [
    "DataTransformNode",
    "DataTransformNodeInput",
    "DataTransformNodeOutput",
    "DataTransformNodeParams",
]



"""Conditional nodes - If and IfElse branching."""

import logging
from collections.abc import Sequence
from functools import cached_property
from typing import Literal

from overrides import override
from pydantic import ConfigDict, Field
from workflow_engine import (
    BooleanValue,
    Context,
    Data,
    Empty,
    JSONValue,
    Params,
)
from workflow_engine.core import NodeTypeInfo as WENodeTypeInfo

from ..field import FieldInfo, FieldType
from ..node_base import AceTeamNode
from ..node_info import NodeTypeInfo
from ..workflow import AceTeamWorkflow

logger = logging.getLogger(__name__)


WORKFLOW_IF_TRUE_FIELD_INFO = FieldInfo(
    name="if_true",
    display_name="Workflow If True",
    type=FieldType.WORKFLOW,
    description="The workflow to run if the condition is true.",
)

WORKFLOW_IF_FALSE_FIELD_INFO = FieldInfo(
    name="if_false",
    display_name="Workflow If False",
    type=FieldType.WORKFLOW,
    description="The workflow to run if the condition is false.",
)

CONDITION_FIELD_INFO = FieldInfo(
    name="condition",
    display_name="Condition",
    type=FieldType.BOOLEAN,
    description="The condition to check.",
)


class IfParams(Params):
    if_true: JSONValue = Field(default_factory=lambda: JSONValue(None))

    @cached_property
    def workflow_if_true(self) -> AceTeamWorkflow:
        return AceTeamWorkflow.model_validate(self.if_true.root)


class IfElseParams(Params):
    if_true: JSONValue = Field(default_factory=lambda: JSONValue(None))
    if_false: JSONValue = Field(default_factory=lambda: JSONValue(None))

    @cached_property
    def workflow_if_true(self) -> AceTeamWorkflow:
        return AceTeamWorkflow.model_validate(self.if_true.root)

    @cached_property
    def workflow_if_false(self) -> AceTeamWorkflow:
        return AceTeamWorkflow.model_validate(self.if_false.root)


class ConditionalInput(Data):
    model_config = ConfigDict(extra="allow")
    condition: BooleanValue


class IfNode(
    AceTeamNode[Data, Empty, IfParams]
):
    """Runs a workflow only if a boolean condition is true."""

    TYPE_INFO = WENodeTypeInfo.from_parameter_type(
        name="If",
        display_name="If",
        description="Runs a workflow only if a boolean condition is true.",
        version="0.4.0",
        parameter_type=IfParams,
    )

    type: Literal["If"] = "If"  # pyright: ignore[reportIncompatibleVariableOverride]

    @classmethod
    def type_info(cls) -> NodeTypeInfo:
        return NodeTypeInfo(
            type="If",
            display_name="If",
            description="Runs a workflow only if a boolean condition is true.",
            params=(WORKFLOW_IF_TRUE_FIELD_INFO,),
        )

    @cached_property
    def input_fields_info(self) -> Sequence[FieldInfo]:
        true_input_types_by_name = {
            input.name: input.type for input in self.params.workflow_if_true.inputs
        }
        assert "condition" not in true_input_types_by_name, (
            f"{self.type}: condition is a reserved keyword"
        )
        return (CONDITION_FIELD_INFO, *self.params.workflow_if_true.inputs)

    @cached_property
    def output_fields_info(self) -> Sequence[FieldInfo]:
        return self.params.workflow_if_true.outputs

    @override
    async def run(self, context: Context, input: ConditionalInput) -> AceTeamWorkflow | None:
        if input.condition:
            return self.params.workflow_if_true
        else:
            return None


class IfElseNode(
    AceTeamNode[Data, Empty, IfElseParams]
):
    """Runs one of two workflows based on a boolean condition."""

    TYPE_INFO = WENodeTypeInfo.from_parameter_type(
        name="IfElse",
        display_name="If Else",
        description="Runs one of two workflows based on a boolean condition.",
        version="0.4.0",
        parameter_type=IfElseParams,
    )

    type: Literal["IfElse"] = "IfElse"  # pyright: ignore[reportIncompatibleVariableOverride]

    @classmethod
    def type_info(cls) -> NodeTypeInfo:
        return NodeTypeInfo(
            type="IfElse",
            display_name="If Else",
            description="Runs one of two workflows based on a boolean condition.",
            params=(WORKFLOW_IF_TRUE_FIELD_INFO, WORKFLOW_IF_FALSE_FIELD_INFO),
        )

    @cached_property
    def input_fields_info(self) -> Sequence[FieldInfo]:
        true_input_types_by_name = {
            input.name: input.type for input in self.params.workflow_if_true.inputs
        }
        false_input_types_by_name = {
            input.name: input.type for input in self.params.workflow_if_false.inputs
        }
        assert true_input_types_by_name == false_input_types_by_name, (
            f"{self.type}: true and false inputs must match"
        )
        assert "condition" not in true_input_types_by_name, (
            f"{self.type}: condition is a reserved keyword"
        )
        return (CONDITION_FIELD_INFO, *self.params.workflow_if_true.inputs)

    @cached_property
    def output_fields_info(self) -> Sequence[FieldInfo]:
        true_output_types_by_name = {
            output.name: output.type for output in self.params.workflow_if_true.outputs
        }
        false_output_types_by_name = {
            output.name: output.type for output in self.params.workflow_if_false.outputs
        }
        assert true_output_types_by_name == false_output_types_by_name, (
            f"{self.type}: true and false outputs must match"
        )
        return self.params.workflow_if_true.outputs

    @override
    async def run(self, context: Context, input: ConditionalInput) -> AceTeamWorkflow:
        if input.condition:
            return self.params.workflow_if_true
        else:
            return self.params.workflow_if_false


__all__ = [
    "IfElseNode",
    "IfElseParams",
    "IfNode",
    "IfParams",
]



"""Node type info for AceTeam nodes."""

from collections.abc import Sequence

from pydantic import BaseModel

from .field import FieldInfo


class NodeTypeInfo(BaseModel):
    """Represents a node type's metadata."""

    type: str
    display_name: str
    description: str
    params: Sequence[FieldInfo]


__all__ = [
    "NodeTypeInfo",
]



"""AceTeam workflow nodes.

Node submodules are NOT auto-imported here to avoid registration conflicts
when this package is used as a library alongside a platform that defines
its own versions of some nodes (e.g., APICallNode, LLMNode).

Import specific node modules directly:
    from aceteam_nodes.nodes.comparison import EqualNode
    from aceteam_nodes.nodes.conditional import IfNode

Or call register_all_nodes() to trigger auto-registration of all nodes
(e.g., before deserializing a workflow file).
"""


def register_all_nodes():
    """Import all node modules to trigger auto-registration in workflow_engine."""
    from . import api_call as _  # noqa: F401
    from . import comparison as _  # noqa: F811, F401
    from . import conditional as _  # noqa: F811, F401
    from . import csv_reader as _  # noqa: F811, F401
    from . import data_transform as _  # noqa: F811, F401
    from . import iteration as _  # noqa: F811, F401
    from . import llm as _  # noqa: F811, F401
    from . import text_io as _  # noqa: F811, F401



"""Field types and field info models for AceTeam nodes."""

from __future__ import annotations

from collections.abc import Mapping, Sequence
from csv import DictReader, DictWriter
from enum import StrEnum
from functools import cached_property
from io import StringIO
from typing import Any, ClassVar, Literal, Type, TypeVar

from pydantic import BaseModel, create_model, model_validator
from workflow_engine import (
    BooleanValue,
    Context,
    Data,
    FileValue,
    FloatValue,
    JSONValue,
    SequenceValue,
    StringValue,
    Value,
    ValueType,
)
from workflow_engine.core.values.value import value_type_registry
from workflow_engine.files import (
    JSONFileValue,
    JSONLinesFileValue,
    PDFFileValue,
    TextFileValue,
)


# Value subclasses auto-register in workflow_engine's value type registry.
# When used alongside a platform that defines the same types (e.g., aceteam-8),
# we reuse the already-registered class to avoid duplicate registration errors.


def _get_or_define_value_type(name: str, cls_factory):
    """Return existing registered class or define a new one."""
    if name in value_type_registry.types:
        return value_type_registry.types[name]
    return cls_factory()


def _make_csv_file_value():
    class CSVFileValue(TextFileValue):
        """A CSV file."""

        mime_type: ClassVar[str] = "text/csv"

        async def read_data(self, context: Context) -> Sequence[Mapping[str, Any]]:
            text = await self.read_text(context)
            text_io = StringIO(text)
            reader = DictReader(text_io)
            return tuple(reader)

        async def write_data(self, context: Context, data: Sequence[Mapping[str, Any]]):
            text_io = StringIO()
            writer = DictWriter(text_io, fieldnames=data[0].keys())
            writer.writeheader()
            writer.writerows(data)
            text = text_io.getvalue()
            return await self.write_text(context, text)

    return CSVFileValue


CSVFileValue = _get_or_define_value_type("CSVFileValue", _make_csv_file_value)


class FieldType(StrEnum):
    """The set of field types supported by AceTeam nodes."""

    BOOLEAN = "boolean"
    FIELD_LIST = "field_list"
    ITERATION_INDEXED_FIELD_LIST = "iteration_indexed_field_list"
    FILE = "file"
    FILE_AUDIO = "file:audio"
    FILE_CSV = "file:csv"
    FILE_JSON = "file:json"
    FILE_JSONLINES = "file:jsonlines"
    FILE_MP3 = "file:mp3"
    FILE_PDF = "file:pdf"
    FILE_TEXT = "file:text"
    FILE_VIDEO = "file:video"
    GANTT_CHART = "gantt_chart"
    GRAPH = "graph"
    JSON = "json"
    LONG_TEXT = "long_text"
    MODEL = "model"
    MODEL_LIST = "model_list"
    NUMBER = "number"
    SEQUENCE = "sequence"
    SHORT_TEXT = "short_text"
    SQL = "sql"
    WORKFLOW = "workflow"


class FieldInfo(BaseModel):
    """Represents metadata for a node field."""

    name: str
    type: FieldType
    display_name: str | None = None
    description: str | None = None
    default: Any | None = None
    type_parameters: Sequence[FieldInfo] = ()

    # specific to number fields
    min: float | None = None
    max: float | None = None
    step: float | None = None

    # specific to SQL fields
    database_connection: str | None = None

    # specific to template fields
    template_parameters: str | None = None

    # specific to text fields - list of options for dropdown/select
    options: Sequence[str] | None = None

    @model_validator(mode="after")
    def _validate_number_fields(self):
        if self.type is not FieldType.NUMBER:
            assert self.min is None
            assert self.max is None
            assert self.step is None

        if self.type is not FieldType.SQL:
            assert self.database_connection is None

        if self.type not in (FieldType.LONG_TEXT, FieldType.SHORT_TEXT):
            assert self.template_parameters is None

        if self.type != FieldType.SHORT_TEXT:
            assert self.options is None

        if self.min is not None and self.max is not None:
            assert self.min <= self.max
        if self.min is not None and self.default is not None:
            assert self.min <= self.default
        if self.max is not None and self.default is not None:
            assert self.default <= self.max

        return self

    @model_validator(mode="after")
    def _validate_sequence_fields(self):
        if self.type is FieldType.SEQUENCE:
            assert len(self.type_parameters) <= 1
        else:
            assert len(self.type_parameters) == 0
        return self

    @property
    def python_type(self) -> ValueType:
        if self.type is FieldType.SEQUENCE:
            if self.type_parameters:
                (element_type,) = self.type_parameters
                return SequenceValue[element_type.python_type]
            return SequenceValue[JSONValue]

        return _non_parametric_field_type_to_type[self.type]


def _make_field_info_value():
    class FieldInfoValue(Value[FieldInfo]):
        @property
        def name(self) -> str:
            return self.root.name

        @property
        def type(self) -> FieldType:
            return self.root.type

        @property
        def display_name(self) -> str | None:
            return self.root.display_name

        @property
        def description(self) -> str | None:
            return self.root.description

        @cached_property
        def default(self) -> Value | None:
            if self.root.default is None:
                return None
            if isinstance(self.root.default, self.python_type):
                return self.root.default
            return self.python_type.model_validate(self.root.default)

        @property
        def min(self) -> float | None:
            return self.root.min

        @property
        def max(self) -> float | None:
            return self.root.max

        @property
        def step(self) -> float | None:
            return self.root.step

        @property
        def database_connection(self) -> str | None:
            return self.root.database_connection

        @property
        def template_parameters(self) -> str | None:
            return self.root.template_parameters

        @property
        def python_type(self) -> ValueType:
            return self.root.python_type

        @property
        def options(self) -> Sequence[str] | None:
            return self.root.options

    return FieldInfoValue


FieldInfoValue = _get_or_define_value_type("FieldInfoValue", _make_field_info_value)


class IterationIndexedFieldInfo(FieldInfo):
    """A field that may be indexed by an iteration index."""

    iteration_index: int | None = None
    expand: Literal[False] = False


def _make_iteration_indexed_field_info_value():
    class IterationIndexedFieldInfoValue(Value[IterationIndexedFieldInfo]):
        @property
        def name(self) -> str:
            return self.root.name

        @property
        def type(self) -> FieldType:
            return self.root.type

        @property
        def display_name(self) -> str | None:
            return self.root.display_name

        @property
        def description(self) -> str | None:
            return self.root.description

        @property
        def default(self) -> Any | None:
            return self.root.default

        @property
        def min(self) -> float | None:
            return self.root.min

        @property
        def max(self) -> float | None:
            return self.root.max

        @property
        def step(self) -> float | None:
            return self.root.step

        @property
        def database_connection(self) -> str | None:
            return self.root.database_connection

        @property
        def template_parameters(self) -> str | None:
            return self.root.template_parameters

        @property
        def python_type(self) -> ValueType:
            return self.root.python_type

        @property
        def options(self) -> Sequence[str] | None:
            return self.root.options

        @property
        def iteration_index(self) -> int | None:
            return self.root.iteration_index

        def to_field_info(self) -> FieldInfo:
            return FieldInfo(
                name=self.name,
                type=self.type,
                display_name=self.display_name,
                description=self.description,
                default=self.default,
                min=self.min,
                max=self.max,
                step=self.step,
                database_connection=self.database_connection,
                template_parameters=self.template_parameters,
                options=self.options,
            )

    return IterationIndexedFieldInfoValue


IterationIndexedFieldInfoValue = _get_or_define_value_type(
    "IterationIndexedFieldInfoValue", _make_iteration_indexed_field_info_value
)


type FieldSequenceValue = SequenceValue[FieldInfoValue]
type IterationIndexedFieldSequenceValue = SequenceValue[IterationIndexedFieldInfoValue]


D = TypeVar("D", bound=Data)


def build_pydantic_model(
    fields_info: Sequence[FieldInfo | FieldInfoValue] | FieldSequenceValue,
    base_cls: Type[D] = Data,
) -> Type[D]:
    if len(fields_info) == 0:
        return base_cls
    if isinstance(fields_info, SequenceValue):
        fields_info = fields_info.root
    return create_model(
        "Input",
        __base__=base_cls,
        **{info.name: info.python_type for info in fields_info},  # type: ignore
    )


_non_parametric_field_type_to_type: Mapping[FieldType, ValueType] = {
    FieldType.BOOLEAN: BooleanValue,
    FieldType.FIELD_LIST: SequenceValue[FieldInfoValue],
    FieldType.ITERATION_INDEXED_FIELD_LIST: SequenceValue[IterationIndexedFieldInfoValue],
    FieldType.FILE: FileValue,
    FieldType.FILE_AUDIO: FileValue,
    FieldType.FILE_CSV: CSVFileValue,
    FieldType.FILE_JSON: JSONFileValue,
    FieldType.FILE_JSONLINES: JSONLinesFileValue,
    FieldType.FILE_MP3: FileValue,
    FieldType.FILE_PDF: PDFFileValue,
    FieldType.FILE_TEXT: TextFileValue,
    FieldType.FILE_VIDEO: FileValue,
    FieldType.GANTT_CHART: JSONValue,
    FieldType.GRAPH: JSONValue,
    FieldType.LONG_TEXT: StringValue,
    FieldType.JSON: JSONValue,
    FieldType.MODEL: StringValue,
    FieldType.MODEL_LIST: SequenceValue[StringValue],
    FieldType.NUMBER: FloatValue,
    FieldType.SHORT_TEXT: StringValue,
    FieldType.SQL: StringValue,
    FieldType.WORKFLOW: JSONValue,
}


__all__ = [
    "CSVFileValue",
    "FieldInfo",
    "FieldInfoValue",
    "FieldSequenceValue",
    "FieldType",
    "IterationIndexedFieldInfo",
    "IterationIndexedFieldInfoValue",
    "IterationIndexedFieldSequenceValue",
    "build_pydantic_model",
]



"""CSV Reader node - reads CSV data."""

from collections.abc import Sequence
from typing import Literal

from overrides import override
from pydantic import Field
from workflow_engine import Context, Data, Params, StringValue
from workflow_engine.core import NodeTypeInfo as WENodeTypeInfo

from ..field import FieldInfo, FieldType
from ..node_base import AceTeamNode
from ..node_info import NodeTypeInfo


class CSVReaderNodeParams(Params):
    sample_data: StringValue = Field(
        default=StringValue("name,value\nexample,1"),
        description="Embedded CSV data",
    )


class CSVReaderNodeOutput(Data):
    data: StringValue


class CSVReaderNode(
    AceTeamNode[None, CSVReaderNodeOutput, CSVReaderNodeParams],
):
    """Reads CSV data and outputs it as text."""

    TYPE_INFO = WENodeTypeInfo.from_parameter_type(
        name="CSVReader",
        display_name="CSV Reader",
        description="Reads CSV data and outputs it as text.",
        version="1.0.0",
        parameter_type=CSVReaderNodeParams,
    )

    type: Literal["CSVReader"] = "CSVReader"

    @classmethod
    @override
    def type_info(cls) -> NodeTypeInfo:
        return NodeTypeInfo(
            type="CSVReader",
            display_name="CSV Reader",
            description="Reads CSV data and outputs it as text.",
            params=(
                FieldInfo(
                    name="sample_data",
                    display_name="Sample Data",
                    description="Embedded CSV data",
                    type=FieldType.LONG_TEXT,
                    default="name,value\nexample,1",
                ),
            ),
        )

    @property
    def input_fields_info(self) -> Sequence[FieldInfo]:
        return ()

    @property
    def output_fields_info(self) -> Sequence[FieldInfo]:
        return (
            FieldInfo(
                name="data",
                display_name="Data",
                description="The CSV data as text",
                type=FieldType.LONG_TEXT,
            ),
        )

    @override
    async def run(self, context: Context, input: None) -> CSVReaderNodeOutput:
        return CSVReaderNodeOutput(data=self.params.sample_data)


__all__ = [
    "CSVReaderNode",
    "CSVReaderNodeOutput",
    "CSVReaderNodeParams",
]



"""AceTeam workflow nodes for local execution."""

__version__ = "0.1.2"

from .field import FieldInfo, FieldType
from .node_base import AceTeamNode, aceteam_nodes
from .node_info import NodeTypeInfo
from .workflow import AceTeamWorkflow


def __getattr__(name: str):
    """Lazy imports for modules that require optional dependencies (litellm)."""
    if name == "CLIContext":
        from .context import CLIContext

        return CLIContext
    if name == "run_workflow_from_file":
        from .execution import run_workflow_from_file

        return run_workflow_from_file
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


__all__ = [
    "AceTeamNode",
    "AceTeamWorkflow",
    "CLIContext",
    "FieldInfo",
    "FieldType",
    "NodeTypeInfo",
    "aceteam_nodes",
    "run_workflow_from_file",
]
